import threading
from typing import Any, Optional

from . import futures
from .async_task import AsyncTask

class AsyncTaskError(Exception): ...

class TaskScheduler(object):
    def __init__(self) -> None: ...
    def reset(self) -> None: ...
    def wait_for(self, task: futures.FutureBase[Any]) -> None: ...
    def dump(self, indent: int = ...) -> None: ...
    def try_time_based_dump(self, last_task: Any = ...) -> None: ...

class LocalTaskSchedulerState(threading.local):
    def reset(self) -> None: ...

def get_scheduler() -> TaskScheduler: ...
def reset() -> None: ...
def get_active_task() -> Optional[futures.FutureBase]: ...



import traceback
from typing import Callable, Generic, NoReturn, Optional, TypeVar

import qcore.helpers as core_helpers
import qcore.events as core_events
import qcore.errors as core_errors

from . import debug
from . import _debug

_T = TypeVar("_T")

class FutureIsAlreadyComputed(Exception): ...

class FutureBase(Generic[_T]):
    def __init__(self) -> None: ...
    def value(self) -> _T: ...
    def set_value(self, value: _T) -> None: ...
    def reset_unsafe(self) -> None: ...
    def error(self) -> Optional[BaseException]: ...
    def set_error(self, error: BaseException) -> None: ...
    def is_computed(self) -> bool: ...
    def raise_if_error(self) -> None: ...
    def __call__(self) -> _T: ...
    def dump(self, indent: int = ...) -> None: ...

class Future(FutureBase[_T]):
    def __init__(self, value_provider: Callable[..., _T]) -> None: ...

class ConstFuture(FutureBase[_T]):
    def __init__(self, value: _T) -> None: ...

none_future: ConstFuture[None]

class ErrorFuture(FutureBase[NoReturn]):
    def __init__(self, error: BaseException) -> None: ...



from typing import overload

class DebugOptions(object):
    DUMP_PRE_ERROR_STATE: bool
    DUMP_EXCEPTIONS: bool
    DUMP_SCHEDULE_TASK: bool
    DUMP_CONTINUE_TASK: bool
    DUMP_SCHEDULE_BATCH: bool
    DUMP_FLUSH_BATCH: bool
    DUMP_DEPENDENCIES: bool
    DUMP_COMPUTED: bool
    DUMP_NEW_TASKS: bool
    DUMP_YIELD_RESULTS: bool
    DUMP_QUEUED_RESULTS: bool
    DUMP_CONTEXTS: bool
    DUMP_SYNC: bool
    DUMP_STACK: bool
    DUMP_SCHEDULER_STATE: bool
    DUMP_SYNC_CALLS: bool
    COLLECT_PERF_STATS: bool

    SCHEDULER_STATE_DUMP_INTERVAL: int
    DEBUG_STR_REPR_MAX_LENGTH: int
    STACK_DUMP_LIMIT: int

    ENABLE_COMPLEX_ASSERTIONS: bool
    def __init__(self) -> None: ...
    @overload
    def DUMP_ALL(self, value: None = ...) -> bool: ...
    @overload
    def DUMP_ALL(self, value: bool) -> None: ...

options: DebugOptions



from . import debug
from . import futures
from . import batching
from . import async_task
from . import scheduler
from . import decorators
from . import utils
from . import contexts
from . import scoped_value
from . import tools
from . import profiler
from .futures import (
    FutureBase as FutureBase,
    Future as Future,
    FutureIsAlreadyComputed as FutureIsAlreadyComputed,
    none_future as none_future,
    ConstFuture as ConstFuture,
    ErrorFuture as ErrorFuture,
)
from .batching import (
    BatchBase as BatchBase,
    BatchItemBase as BatchItemBase,
    BatchingError as BatchingError,
    BatchCancelledError as BatchCancelledError,
)
from .async_task import (
    AsyncTask as AsyncTask,
    AsyncTaskCancelledError as AsyncTaskCancelledError,
    AsyncTaskResult as AsyncTaskResult,
)
from .scheduler import (
    TaskScheduler as TaskScheduler,
    get_scheduler as get_scheduler,
    get_active_task as get_active_task,
    AsyncTaskError as AsyncTaskError,
)
from .decorators import (
    asynq as asynq,
    async_proxy as async_proxy,
    has_async_fn as has_async_fn,
    is_pure_async_fn as is_pure_async_fn,
    is_async_fn as is_async_fn,
    get_async_fn as get_async_fn,
    get_async_or_sync_fn as get_async_or_sync_fn,
    async_call as async_call,
    make_async_decorator as make_async_decorator,
)
from .utils import result as result
from .contexts import NonAsyncContext as NonAsyncContext, AsyncContext as AsyncContext
from .scoped_value import (
    AsyncScopedValue as AsyncScopedValue,
    async_override as async_override,
)
from . import mock_ as mock
from .generator import (
    END_OF_GENERATOR as END_OF_GENERATOR,
    async_generator as async_generator,
    list_of_generator as list_of_generator,
    take_first as take_first,
    Value as Value,
)



from builtins import str as _str
import contextlib
import logging
from types import FrameType, GeneratorType, TracebackType
from typing import (
    Any,
    ContextManager,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    overload,
    Union,
)

from ._debug import options as options

@overload
def DUMP_ALL(self, value: None = ...) -> bool: ...
@overload
def DUMP_ALL(self, value: bool) -> None: ...
def enable_original_exc_handler(enable: bool) -> None: ...
def enable_filter_traceback(enable: bool) -> None: ...
def enable_traceback_syntax_highlight(enable: bool) -> None: ...
def dump_error(error: BaseException, tb: Optional[TracebackType] = ...) -> None: ...
def format_error(error: BaseException, tb: Optional[TracebackType] = ...) -> _str: ...

class AsynqStackTracebackFormatter(logging.Formatter):
    def formatException(
        self,
        exc_info: Union[
            Tuple[type, BaseException, Optional[TracebackType]], Tuple[None, None, None]
        ],
    ) -> _str: ...

def extract_tb(
    tb: TracebackType, limit: Optional[int] = ...
) -> List[Tuple[_str, int, _str, _str]]: ...
def format_tb(tb: TracebackType) -> _str: ...
def dump_stack(skip: int = ..., limit: Optional[int] = ...) -> None: ...
def dump_asynq_stack() -> None: ...
def format_asynq_stack() -> Optional[List[_str]]: ...
def dump(state: Any) -> None: ...
def write(text: _str, indent: int = ...) -> None: ...
def str(source: object, truncate: bool = ...) -> _str: ...
def repr(source: object, truncate: bool = ...) -> _str: ...
def async_exception_hook(
    type: Type[BaseException], error: BaseException, tb: TracebackType
) -> None: ...
def ipython_custom_exception_handler(
    self: Any,
    etype: Type[BaseException],
    value: BaseException,
    tb: TracebackType,
    tb_offset: Any = ...,
) -> None: ...
def attach_exception_hook() -> None: ...
def detach_exception_hook() -> None: ...
@contextlib.contextmanager
def enable_complex_assertions(enable: bool = ...) -> Iterator[None]: ...
def disable_complex_assertions() -> ContextManager[None]: ...
def sync() -> None: ...
def get_frame(generator: GeneratorType) -> Optional[FrameType]: ...
def filter_traceback(tb_list: List[_str]) -> List[_str]: ...
def syntax_highlight_tb(_str) -> _str: ...



import random
import threading
import time
import collections

from qcore import SECOND
from qcore.asserts import assert_eq, assert_le, assert_ge, assert_unordered_list_eq
from asynq import asynq, debug, profiler
from .debug_cache import reset_caches, mc
from .caching import ExternalCacheBatchItem


class ProfilerThread(threading.Thread):
    def __init__(self, stats_list):
        super().__init__()
        self.stats_list = stats_list[:]

    def run(self):
        profiler.reset()
        for stats in self.stats_list:
            profiler.append(stats)
        self.profiled_result = profiler.flush()


def test_multithreaded_profiler():
    """Test multithreaded profiler."""
    stats_list = list(range(10000))

    profiler.reset()
    assert_eq([], profiler.flush())

    threads = []
    for i in range(30):
        random.shuffle(stats_list)
        thread = ProfilerThread(stats_list)
        threads.append(thread)

    random.shuffle(stats_list)
    for stats in stats_list:
        profiler.append(stats)

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()
        assert_eq(thread.stats_list, thread.profiled_result)

    assert_eq(stats_list, profiler.flush())


def test_collect_perf_stats():
    sleep_time = 50000  # 50ms

    @asynq()
    def func(depth):
        if depth == 0:
            return (yield ExternalCacheBatchItem(mc._batch, "get", "test"))
        time.sleep(float(sleep_time) / SECOND)
        yield func.asynq(depth - 1), func.asynq(depth - 1)

    debug.options.COLLECT_PERF_STATS = True
    debug.options.KEEP_DEPENDENCIES = True
    profiler.reset()
    reset_caches()
    depth = 3
    func(depth)
    profiled_result = profiler.flush()

    batches = 1
    async_tasks = 2 ** (depth + 1) - 1
    num_leaf_tasks = 2**depth

    assert_eq(async_tasks + batches, len(profiled_result))

    deps_to_tasks = collections.defaultdict(list)
    for task in profiled_result:
        num_deps = len(task["dependencies"])
        deps_to_tasks[num_deps].append(task)

    assert_unordered_list_eq([1, 2, num_leaf_tasks], deps_to_tasks.keys())

    # leaf tasks
    assert_eq(num_leaf_tasks, len(deps_to_tasks[1]))
    for task in deps_to_tasks[1]:
        assert_ge(sleep_time, task["time_taken"])

    # one batch (with all the batch items as dependencies)
    assert_eq(1, len(deps_to_tasks[num_leaf_tasks]))

    # non-leaf async tasks in the tree (1 + 2 + 3 = 7)
    assert_eq(7, len(deps_to_tasks[2]))
    for task in deps_to_tasks[2]:
        assert_le(sleep_time, task["time_taken"])

    # When COLLET_PERF_STATS is False, we shouldn't profile anything.
    debug.options.COLLECT_PERF_STATS = False
    profiler.reset()
    reset_caches()
    func(2)
    profiled_result = profiler.flush()
    assert_eq(0, len(profiled_result))



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import asynq, result, debug
from qcore.asserts import AssertRaises


@asynq()
def bad_function():
    yield bad_function.asynq()


def test_it():
    # Adjust max stack size to make test take less memory and time
    default_stack_limit = debug.options.MAX_TASK_STACK_SIZE
    try:
        debug.options.MAX_TASK_STACK_SIZE = 200
        with AssertRaises(RuntimeError):
            bad_function()
    finally:
        debug.options.MAX_TASK_STACK_SIZE = default_stack_limit



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asynq import asynq
from asynq.futures import ConstFuture
from asynq.contexts import AsyncContext
from asynq import scheduler

from qcore.asserts import assert_eq, assert_is, AssertRaises


def check_unwrap(expected, to_unwrap):
    # can't call unwrap() directly because it's hidden by cython
    @asynq()
    def caller():
        value = yield to_unwrap
        assert_eq(expected, value, extra="yielded {}".format(to_unwrap))

    caller()


def test_unwrap():
    check_unwrap(None, None)
    check_unwrap(1, ConstFuture(1))

    for typ in (list, tuple):
        check_unwrap(typ(), typ())
        check_unwrap(typ([1]), typ([ConstFuture(1)]))
        check_unwrap(typ([1, 2]), typ([ConstFuture(1), ConstFuture(2)]))
        check_unwrap(
            typ([1, 2, 3]), typ([ConstFuture(1), ConstFuture(2), ConstFuture(3)])
        )

    with AssertRaises(TypeError):
        check_unwrap(1)


class Ctx(AsyncContext):
    def __init__(self, ctx_id):
        self.ctx_id = ctx_id

    def __eq__(self, other):
        return self.ctx_id == other.ctx_id

    def resume(self):
        pass

    def pause(self):
        pass


def test_context_nesting():
    """Make sure that context nesting ordering is preserved.

    This should be true even when two contexts are __eq__ to one another.

    """

    @asynq()
    def fn():
        task = scheduler.get_active_task()
        ctx0 = Ctx(0)
        ctx1 = Ctx(1)
        ctx2 = Ctx(0)
        with ctx0:
            with ctx1:
                with ctx2:
                    assert_eq([ctx0, ctx1, ctx2], list(task._contexts.values()))

                assert_eq([ctx0, ctx1], list(task._contexts.values()))

    fn()


def test_clear_dependencies():
    """Ensure that we the list of dependencies won't continue growing for a long-running task."""

    @asynq()
    def inner_fn():
        return None

    @asynq()
    def fn():
        task = scheduler.get_active_task()
        assert_eq(0, len(task._dependencies))
        yield [inner_fn.asynq() for _ in range(10)]
        assert_eq(0, len(task._dependencies))

    fn()



import types
from typing import Any, Callable, Iterable, List, Mapping, TypeVar

from . import futures

import asynq

_T = TypeVar("_T")

class AsyncTaskCancelledError(GeneratorExit): ...

class AsyncTaskResult(GeneratorExit):
    def __init__(self, result: Any) -> None: ...

class AsyncTask(futures.FutureBase[_T]):
    def __init__(
        self,
        generator: types.GeneratorType,
        fn: Callable[..., Any],
        args: Iterable[object],
        kwargs: Mapping[str, object],
    ) -> None: ...
    def can_continue(self) -> bool: ...
    def is_blocked(self) -> bool: ...
    def dump_perf_stats(self) -> None: ...
    def to_str(self) -> str: ...
    def collect_perf_stats(self) -> None: ...
    def traceback(self) -> List[str]: ...
    def dump(self, indent: int = ...) -> None: ...



import threading
from typing import Any, Generic, List, Optional, Tuple, TypeVar

from . import futures

_T = TypeVar("_T", bound=BatchItemBase)

class BatchingError(Exception): ...
class BatchCancelledError(BatchingError): ...

class BatchBase(futures.FutureBase, Generic[_T]):
    items: List[_T]
    def __init__(self) -> None: ...
    def is_flushed(self) -> bool: ...
    def is_cancelled(self) -> bool: ...
    def is_empty(self) -> bool: ...
    def get_priority(self) -> Tuple[int, int]: ...
    def flush(self) -> None: ...
    def cancel(self, error: Optional[Exception] = ...) -> None: ...
    def dump(self, indent: int = ...) -> None: ...

class BatchItemBase(futures.FutureBase):
    batch: BatchBase
    index: int
    def __init__(self: _T, batch: BatchBase[_T]) -> None: ...

class DebugBatchItem(BatchItemBase):
    def __init__(self, batch_name: str = ..., result: Any = ...) -> None: ...

class DebugBatch(BatchBase):
    def __init__(self, name: str = ..., index: int = 0) -> None: ...

class LocalDebugBatchState(threading.local): ...

def sync(tag: str = ...) -> DebugBatchItem: ...



# Copyright 2016 Quora, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import traceback
import logging

from io import StringIO
from qcore.asserts import assert_eq, assert_in, assert_is, assert_is_not
from qcore import MarkerObject, prepare_for_reraise

import asynq


def test_dump_error():
    buf = StringIO()

    with asynq.mock.patch("asynq.debug.stderr", buf):
        asynq.debug.dump_error(None)
    assert_eq("\nNo error", buf.getvalue())


@asynq.asynq()
def async_fn():
    pass


def test_format_error():
    assert_is(None, asynq.debug.format_error(None))

    # Syntax highlighting adds color text between words
    asynq.debug.enable_traceback_syntax_highlight(False)
    e = RuntimeError()
    expected = "RuntimeError\n"
    assert_eq(expected, asynq.debug.format_error(e))

    e._task = async_fn.asynq()
    formatted = asynq.debug.format_error(e)
    assert_in(expected, formatted)

    try:
        raise RuntimeError
    except RuntimeError:
        e._traceback = sys.exc_info()[2]

    formatted = asynq.debug.format_error(e)
    assert_in(expected, formatted)
    assert_in("Traceback", formatted)

    # Each single word, and unformatted text should be present
    asynq.debug.enable_traceback_syntax_highlight(True)

    expected = "RuntimeError"
    formatted = asynq.debug.format_error(e)
    assert_in(expected, formatted)
    assert_in("Traceback", formatted)


def test_format_error_chaining():
    try:
        try:
            raise ValueError
        except ValueError:
            raise KeyError
    except KeyError as e:
        prepare_for_reraise(e)
        exc = e
    # Syntax highlighting adds color text between words
    asynq.debug.enable_traceback_syntax_highlight(False)
    formatted = asynq.debug.format_error(exc)
    assert_in("raise ValueError", formatted)
    assert_in("raise KeyError", formatted)
    assert_in("During handling of the", formatted)

    # Each single word, and unformatted text should be present
    asynq.debug.enable_traceback_syntax_highlight(True)
    formatted = asynq.debug.format_error(exc)
    assert_in("ValueError", formatted)
    assert_in("KeyError", formatted)
    assert_in("During handling of the", formatted)


def test_dump_stack():
    buf = StringIO()

    with asynq.mock.patch("sys.stdout", buf):

        def inner():
            asynq.debug.dump_stack()

        inner()

    printed = buf.getvalue()
    assert_in("test_dump_stack", printed)
    assert_in("Stack trace:", printed)


def test_format_asynq_stack():
    format_list = []

    @asynq.asynq()
    def level1(arg):
        if arg == 0:
            format_list.append(asynq.debug.format_asynq_stack())
        return arg

    @asynq.asynq()
    def level2(arg):
        return (yield level1.asynq(arg))

    @asynq.asynq()
    def root():
        vals = yield [level2.asynq(i) for i in range(5)]
        return vals

    root()

    traceback = "\n".join(format_list[0])
    assert_in("root", traceback)
    assert_in("level1", traceback)
    assert_in("level2", traceback)


def _assert_write_result(text, indent, expected):
    buf = StringIO()

    with asynq.mock.patch("asynq.debug.stdout", buf):
        asynq.debug.write(text, indent=indent)

    assert_eq(expected, buf.getvalue())


def test_write():
    _assert_write_result("capybara", 0, "capybara\n")
    _assert_write_result("capybara", 1, "  capybara\n")
    _assert_write_result("capybara", 2, "    capybara\n")
    _assert_write_result("nutria\ncapybara", 1, "  nutria\n  capybara\n")
    _assert_write_result("", 1, "  \n")
    _assert_write_result("\n", 1, "\n  \n")
    _assert_write_result("\ncapybara", 1, "\n  capybara\n")


def non_async_function_that_raises_an_error():
    raise ValueError


@asynq.asynq()
def async_function_that_raises_an_error():
    yield None
    non_async_function_that_raises_an_error()


@asynq.asynq()
def async_function_whose_child_async_task_will_throw_an_error():
    yield async_function_that_raises_an_error.asynq()


def test_asynq_traceback_gets_glued_at_each_task_level():
    # tests that exceptions are producting the right tracebacks
    traceback_to_verify = None
    try:
        async_function_whose_child_async_task_will_throw_an_error()
    except ValueError:
        traceback_to_verify = sys.exc_info()[2]
    assert_is_not(None, traceback_to_verify)
    traceback_printed = "\n".join(traceback.format_tb(traceback_to_verify))
    assert_in(non_async_function_that_raises_an_error.__name__, traceback_printed)
    assert_in(async_function_that_raises_an_error.__name__, traceback_printed)
    assert_in(
        async_function_whose_child_async_task_will_throw_an_error.__name__,
        traceback_printed,
    )


def assert_eq_extracted_traceback_entry(entry, filename, fn_name, line):
    entry_filename, _, entry_fn_name, entry_line = entry
    assert_in(filename.rstrip("c"), entry_filename)
    assert_eq(fn_name, entry_fn_name)
    assert_eq(line, entry_line)


def test_extract_traceback():
    traceback_to_verify = None
    try:
        async_function_whose_child_async_task_will_throw_an_error()
    except ValueError:
        traceback_to_verify = sys.exc_info()[2]
    extracted_traceback_to_verify = asynq.debug.extract_tb(traceback_to_verify)

    this_level = extracted_traceback_to_verify[0]
    assert_eq_extracted_traceback_entry(
        this_level,
        __file__,
        "test_extract_traceback",
        "async_function_whose_child_async_task_will_throw_an_error()",
    )

    # now check the last 3 frames. Between the 1st and the last 3 frames there may
    # or may not be a frame from six.reraise depending on the python version.
    async_wrapper_level = extracted_traceback_to_verify[-3]
    assert_eq_extracted_traceback_entry(
        async_wrapper_level,
        __file__,
        "async_function_whose_child_async_task_will_throw_an_error",
        "yield async_function_that_raises_an_error.asynq()",
    )

    async_raiser_level = extracted_traceback_to_verify[-2]
    assert_eq_extracted_traceback_entry(
        async_raiser_level,
        __file__,
        "async_function_that_raises_an_error",
        "non_async_function_that_raises_an_error()",
    )

    normal_function_raising_exception_level = extracted_traceback_to_verify[-1]
    assert_eq_extracted_traceback_entry(
        normal_function_raising_exception_level,
        __file__,
        "non_async_function_that_raises_an_error",
        "raise ValueError",
    )


a_return_value = MarkerObject("A return value.")


@asynq.mock.patch("asynq.debug.traceback.format_list")
@asynq.mock.patch("asynq.debug.extract_tb")
def test_format_tb(mock_extract_tb, mock_format_list):
    mock_extract_tb.return_value = a_return_value
    mock_format_list.side_effect = lambda arg: arg
    traceback_to_verify = None
    try:
        async_function_whose_child_async_task_will_throw_an_error()
    except ValueError:
        traceback_to_verify = sys.exc_info()[2]
    assert_is(a_return_value, asynq.debug.format_tb(traceback_to_verify))
    mock_extract_tb.assert_called_once_with(traceback_to_verify)
    mock_format_list.assert_called_once_with(a_return_value)


@asynq.mock.patch("asynq.debug.format_error")
def test_asynq_stack_trace_formatter(mock_format_error):
    mock_format_error.return_value = "This is some traceback."
    stderr_string_io = StringIO()
    handler = logging.StreamHandler(stream=stderr_string_io)
    handler.setFormatter(asynq.debug.AsynqStackTracebackFormatter())
    logger = logging.getLogger("test_asynq")
    logger.addHandler(handler)
    exc_info = None
    try:
        async_function_whose_child_async_task_will_throw_an_error()
    except ValueError:
        exc_info = sys.exc_info()
        logger.exception("Test")
    ty, val, tb = exc_info
    mock_format_error.assert_called_once_with(val, tb=tb)
    assert_eq("Test\nThis is some traceback.\n", stderr_string_io.getvalue())


def test_filter_traceback():
    test_replacements = """
  File "asynq/decorators.py", line 161, in asynq.decorators.AsyncDecorator.__call__
  File "asynq/futures.py", line 54, in asynq.futures.FutureBase.value
  File "asynq/futures.py", line 63, in asynq.futures.FutureBase.value
  File "asynq/futures.py", line 153, in asynq.futures.FutureBase.raise_if_error
  File "<...>/python3.6/site-packages/qcore/errors.py", line 93, in reraise
    six.reraise(type(error), error, error._traceback)
  File "<...>/python3.6/site-packages/six.py", line 693, in reraise
    raise value
  File "asynq/async_task.py", line 169, in asynq.async_task.AsyncTask._continue
  File "asynq/async_task.py", line 237, in asynq.async_task.AsyncTask._continue_on_generator
  File "asynq/async_task.py", line 209, in asynq.async_task.AsyncTask._continue_on_generator
  File "asynq/decorators.py", line 153, in asynq.decorators.AsyncDecorator.asynq
  File "asynq/decorators.py", line 203, in asynq.decorators.AsyncProxyDecorator._call_pure
  File "asynq/decorators.py", line 203, in asynq.decorators.AsyncProxyDecorator._call_pure
  File "asynq/decorators.py", line 204, in asynq.decorators.AsyncProxyDecorator._call_pure
  File "asynq/decorators.py", line 275, in asynq.decorators.async_call
  File "something.py", line 25 in hello_world
    hello()
"""

    expected_replacements = [
        "\n",
        "  ___asynq_future_raise_if_error___\n",
        "  ___asynq_continue___\n",
        "  ___asynq_call_pure___\n",
        '  File "something.py", line 25 in hello_world\n',
        "    hello()\n",
    ]

    assert_eq(
        expected_replacements,
        asynq.debug.filter_traceback(test_replacements.splitlines(True)),
    )

    assert_eq([], asynq.debug.filter_traceback([]))

    test_no_replacement = """
  File "something.py", line 25 in hello_world
    hello()
"""

    expected_no_replacement = [
        "\n",
        '  File "something.py", line 25 in hello_world\n',
        "    hello()\n",
    ]
    assert_eq(
        expected_no_replacement,
        asynq.debug.filter_traceback(test_no_replacement.splitlines(True)),
    )

    test_partial_match = """
  File "asynq/decorators.py", line 153, in asynq.decorators.AsyncDecorator.asynq
  File "asynq/decorators.py", line 203, in asynq.decorators.AsyncProxyDecorator._call_pure
  File "asynq/decorators.py", line 203, in asynq.decorators.AsyncProxyDecorator._call_pure
  File "something.py", line 25 in hello_world
    hello()
"""

    # A full match should be required, this is a partial match
    expected_partial_match = [
        "\n",
        (
            '  File "asynq/decorators.py", line 153, in'
            " asynq.decorators.AsyncDecorator.asynq\n"
        ),
        (
            '  File "asynq/decorators.py", line 203, in'
            " asynq.decorators.AsyncProxyDecorator._call_pure\n"
        ),
        (
            '  File "asynq/decorators.py", line 203, in'
            " asynq.decorators.AsyncProxyDecorator._call_pure\n"
        ),
        '  File "something.py", line 25 in hello_world\n',
        "    hello()\n",
    ]

    assert_eq(
        expected_partial_match,
        asynq.debug.filter_traceback(test_partial_match.splitlines(True)),
    )

    test_partial_match_end = """
  File "asynq/decorators.py", line 153, in asynq.decorators.AsyncDecorator.asynq
  File "asynq/decorators.py", line 203, in asynq.decorators.AsyncProxyDecorator._call_pure
  File "asynq/decorators.py", line 203, in asynq.decorators.AsyncProxyDecorator._call_pure
"""

    # A full match should be required, this is a partial match
    expected_partial_match_end = [
        "\n",
        (
            '  File "asynq/decorators.py", line 153, in'
            " asynq.decorators.AsyncDecorator.asynq\n"
        ),
        (
            '  File "asynq/decorators.py", line 203, in'
            " asynq.decorators.AsyncProxyDecorator._call_pure\n"
        ),
        (
            '  File "asynq/decorators.py", line 203, in'
            " asynq.decorators.AsyncProxyDecorator._call_pure\n"
        ),
    ]

    assert_eq(
        expected_partial_match_end,
        asynq.debug.filter_traceback(test_partial_match_end.splitlines(True)),
    )
